
// __________________________________________________________
//
//                         w2k_pdbx.c
//            SBS Program Database Exploder V1.00
//                07-07-2001 Sven B. Schreiber
//                       sbs@orgon.com
// __________________________________________________________

#include "w2k_pdbx.h"

// =================================================================
// DISCLAIMER
// =================================================================

/*

This software is provided "as is" and any express or implied
warranties, including, but not limited to, the implied warranties of
merchantability and fitness for a particular purpose are disclaimed.
In no event shall the author Sven B. Schreiber be liable for any
direct, indirect, incidental, special, exemplary, or consequential
damages (including, but not limited to, procurement of substitute
goods or services; loss of use, data, or profits; or business
interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software,
even if advised of the possibility of such damage.

*/

// =================================================================
// REVISION HISTORY
// =================================================================

/*

07-07-2001 V1.00 Original version (SBS).

*/

// =================================================================
// GLOBAL STRINGS
// =================================================================

WORD awArguments [] = L"{ [+-hard] [/<target>] <PDB path> }";

WORD awMoreInfo  [] = L"\r\n"
                      L"       +   enable subsequent options\r\n"
                      L"       -   disable subsequent options\r\n"
                      L"       h   extract header\r\n"
                      L"       a   extract allocation bits\r\n"
                      L"       r   extract root stream\r\n"
                      L"       d   extract data streams\r\n"
                      L"\r\n"
                      L"Target paths:\r\n"
                      L"\r\n"
                      L"       +h  <target>\\<PDB file>.header\r\n"
                      L"       +a  <target>\\<PDB file>.alloc\r\n"
                      L"       +r  <target>\\<PDB file>.root\r\n"
                      L"       +d  <target>\\<PDB file>.<###>\r\n"
                      L"\r\n"
                      L"       <###> = 0-based stream number.\r\n"
                      L"\r\n"
                      L"If /<target> is omitted, the files are\r\n"
                      L"written to the current directory.\r\n";

// =================================================================
// RUNTIME LIBRARY
// =================================================================

PVOID malloc (size_t size)
    {
    return LocalAlloc (LMEM_FIXED, (size ? size : 1));
    }

// -----------------------------------------------------------------

VOID free (PVOID memblock)
    {
    if (memblock != NULL) LocalFree (memblock);
    return;
    }

// =================================================================
// PDB PROCESSING
// =================================================================

BOOL PdbValid (PPDB_HEADER pph,
               DWORD       dData)
    {
    return (pph != NULL) && (dData >= PDB_HEADER_) &&
           (!lstrcmpA (pph->abSignature, PDB_SIGNATURE_200))&&
           ((DWORD) pph->wFilePages * pph->dPageSize == dData);
    }

// -----------------------------------------------------------------

PPDB_HEADER PdbLoad (PWORD pwPath,
                     PBOOL pfInvalid)
    {
    HANDLE      hf;
    DWORD       dData, dRead;
    PPDB_HEADER pph = NULL;

    *pfInvalid = FALSE;

    if ((hf = CreateFile (pwPath, GENERIC_READ,
                          FILE_SHARE_READ, NULL, OPEN_EXISTING,
                          FILE_FLAG_SEQUENTIAL_SCAN, NULL))
        != INVALID_HANDLE_VALUE)
        {
        dData = GetFileSize (hf, NULL);

        if ((dData != INVALID_FILE_SIZE) && dData &&
            ((pph = malloc (dData)) != NULL))
            {
            if ((!ReadFile (hf, pph, dData, &dRead, NULL)) ||
                (dData != dRead)                           ||
                (*pfInvalid = !PdbValid (pph, dData)))
                {
                free (pph);
                pph = NULL;
                }
            }
        CloseHandle (hf);
        }
    return pph;
    }

// -----------------------------------------------------------------

BOOL PdbSave (PWORD pwTarget,
              PVOID pData,
              DWORD dData)
    {
    HANDLE hf;
    DWORD  dWritten;
    BOOL   fOk = FALSE;

    printf (L"\r\nSaving \"%s\"... ", pwTarget);

    if ((hf = CreateFile (pwTarget, GENERIC_READ | GENERIC_WRITE,
                          FILE_SHARE_READ, NULL, CREATE_ALWAYS,
                          FILE_FLAG_SEQUENTIAL_SCAN |
                          FILE_ATTRIBUTE_NORMAL, NULL))
        != INVALID_HANDLE_VALUE)
        {
        fOk = WriteFile (hf, pData, dData, &dWritten, NULL) &&
              (dData == dWritten);

        CloseHandle (hf);
        }
    printf (fOk ? L"%lu bytes\r\n" : L"ERROR\r\n", dData);
    return fOk;
    }

// -----------------------------------------------------------------

DWORD PdbLimit (PPDB_HEADER pph)
    {
    return ((DWORD) pph->wStartPage - 1) * pph->dPageSize * 8
                                         * pph->dPageSize;
    }

// -----------------------------------------------------------------

DWORD PdbPages (PPDB_HEADER pph,
                DWORD       dBytes)
    {
    return (dBytes ? (((dBytes-1) / pph->dPageSize) + 1) : 0);
    }

// -----------------------------------------------------------------

PVOID PdbRead (PPDB_HEADER pph,
               DWORD       dBytes,
               PWORD       pwPages)
    {
    DWORD i, j;
    DWORD dPages = PdbPages (pph, dBytes);
    PVOID pPages = malloc (dPages * pph->dPageSize);

    if (pPages != NULL)
        {
        for (i = 0; i < dPages; i++)
            {
            j = pwPages [i];

            CopyMemory ((PBYTE) pPages + (i * pph->dPageSize),
                        (PBYTE) pph    + (j * pph->dPageSize),
                        pph->dPageSize);
            }
        }
    return pPages;
    }

// -----------------------------------------------------------------

PPDB_ROOT PdbRoot (PPDB_HEADER pph,
                   PDWORD      pdBytes)
    {
    DWORD     dBytes, i, n;
    PPDB_ROOT ppr = NULL;
    
    if ((ppr = PdbRead (pph, PDB_ROOT_, pph->awRootPages)) != NULL)
        {
        dBytes = PDB_ROOT__ ((DWORD) ppr->wCount);
        free (ppr);

        if ((ppr = PdbRead (pph, dBytes, pph->awRootPages)) != NULL)
            {
            for (n = i = 0; i < (DWORD) ppr->wCount; i++)
                {
                n += PdbPages (pph, ppr->aStreams [i].dStreamSize);
                }
            dBytes += n * sizeof (WORD);
            free (ppr);

            ppr = PdbRead (pph, dBytes, pph->awRootPages);
            }
        }
    *pdBytes = (ppr != NULL ? dBytes : 0);
    return ppr;
    }

// -----------------------------------------------------------------

PVOID PdbStream (PPDB_HEADER pph,
                 PPDB_ROOT   ppr,
                 DWORD       dStream,
                 PDWORD      pdBytes)
    {
    DWORD dBytes, i;
    PWORD pwPages;
    PVOID pPages = NULL;

    if (dStream < (DWORD) ppr->wCount)
        {
        pwPages = (PWORD) ((PBYTE) ppr +
                           PDB_ROOT__ ((DWORD) ppr->wCount));

        for (i = 0; i < dStream; i++)
            {
            pwPages += PdbPages (pph,
                                 ppr->aStreams [i].dStreamSize);
            }
        dBytes = ppr->aStreams [dStream].dStreamSize;
        pPages = PdbRead (pph, dBytes, pwPages);
        }
    *pdBytes = (pPages != NULL ? dBytes : 0);
    return pPages;
    }

// -----------------------------------------------------------------

VOID PdbInfo (PWORD       pwPath,
              PPDB_HEADER pph,
              PPDB_ROOT   ppr)
    {
    DWORD dFileLimit, dFileBytes, dFilePages, dStreams;
    DWORD dDataBytes, dDataPages, dRootBytes, dRootPages, i, n;
    PWORD pwFile;
    WORD  awPath [MAX_PATH];

    dFileLimit = PdbLimit (pph);
    dFilePages = pph->wFilePages;
    dFileBytes = dFilePages * pph->dPageSize;
    dStreams   = ppr->wCount;

    dDataBytes = 0;
    dDataPages = 0;

    for (i = 0; i < dStreams; i++)
        {
        dDataBytes +=                ppr->aStreams [i].dStreamSize ;
        dDataPages += PdbPages (pph, ppr->aStreams [i].dStreamSize);
        }
    dRootBytes = PDB_ROOT__(dStreams) + (dDataPages * sizeof(WORD));
    dRootPages = PdbPages (pph, dRootBytes);

    n = GetFullPathName (pwPath, MAX_PATH, awPath, &pwFile);

    if ((!n) || (n >= MAX_PATH))
        {
        lstrcpyn (awPath, pwPath, MAX_PATH);
        }
    printf (L"\r\n"
            L"Properties of \"%s\":\r\n"
            L"\r\n"
            L"%13lu byte%s maximum size\r\n"
            L"%13lu byte%s allocated\r\n"
            L"%13lu byte%s used by %lu data stream%s\r\n"
            L"%13lu byte%s used by the root stream\r\n"
            L"%13lu byte%s per page\r\n"
            L"%13lu page%s allocated\r\n"
            L"%13lu page%s used by %lu data stream%s\r\n"
            L"%13lu page%s used by the root stream\r\n",
            awPath,
            COUNT (dFileLimit,     L" ", L"s"),
            COUNT (dFileBytes,     L" ", L"s"),
            COUNT (dDataBytes,     L" ", L"s"),
            COUNT (dStreams,       L"",  L"s"),
            COUNT (dRootBytes,     L" ", L"s"),
            COUNT (pph->dPageSize, L" ", L"s"),
            COUNT (dFilePages,     L" ", L"s"),
            COUNT (dDataPages,     L" ", L"s"),
            COUNT (dStreams,       L"",  L"s"),
            COUNT (dRootPages,     L" ", L"s"));
    return;
    }

// -----------------------------------------------------------------

PWORD PdbTarget (PWORD pwPath,
                 PWORD pwTarget,
                 DWORD dOption)
    {
    DWORD i, n;
    BOOL  fSeparator;
    WORD  awExtension [] = L".9876543210";
    PWORD pwExtension    = NULL;
    PWORD pwBuffer       = NULL;

    switch (dOption & DISPLAY_OPTION_FLAGS)
        {
        case DISPLAY_OPTION_HEADER:
            {
            pwExtension = L".header";
            break;
            }
        case DISPLAY_OPTION_ALLOC:
            {
            pwExtension = L".alloc";
            break;
            }
        case DISPLAY_OPTION_ROOT:
            {
            pwExtension = L".root";
            break;
            }
        case DISPLAY_OPTION_DATA:
            {
            wsprintf (awExtension, L".%03lu",
                      dOption & DISPLAY_OPTION_STREAM);

            pwExtension = awExtension;
            break;
            }
        }
    if (pwExtension != NULL)
        {
        i = lstrlen (pwPath);

        while (i && (pwPath [i-1] != '\\')
                 && (pwPath [i-1] != ':' )) i--;

        n = lstrlen (pwTarget);

        if (fSeparator = (n && (pwTarget [n-1] != '\\')
                            && (pwTarget [n-1] != ':' ))) n++;

        n += (lstrlen (pwPath+i) + lstrlen (pwExtension));

        if ((pwBuffer = malloc ((n+1) * sizeof (WORD))) != NULL)
            {
            lstrcpy (pwBuffer, pwTarget);

            if (fSeparator) lstrcat (pwBuffer, L"\\");

            lstrcat (pwBuffer, pwPath+i);
            lstrcat (pwBuffer, pwExtension);
            }
        }
    return pwBuffer;
    }

// -----------------------------------------------------------------

VOID PdbOutput (PWORD       pwPath,
                PWORD       pwTarget,
                DWORD       dOptions,
                PPDB_HEADER pph,
                PPDB_ROOT   ppr,
                DWORD       dRoot)
    {
    PWORD pwTarget1;
    PVOID pData;
    DWORD dData, dStream;

    if (dOptions & DISPLAY_OPTION_HEADER)
        {
        if ((pwTarget1 = PdbTarget (pwPath, pwTarget,
                                    DISPLAY_OPTION_HEADER))
            != NULL)
            {
            PdbSave (pwTarget1, pph, pph->dPageSize);
            free    (pwTarget1);
            }
        else
            {
            printf (L"\r\nUnable to save the header\r\n");
            }
        }
    if (dOptions & DISPLAY_OPTION_ALLOC)
        {
        if ((pwTarget1 = PdbTarget (pwPath, pwTarget,
                                    DISPLAY_OPTION_ALLOC))
            != NULL)
            {
            pData =  (PBYTE) pph + pph->dPageSize;
            dData = ((DWORD) pph->wStartPage - 1) * pph->dPageSize;

            PdbSave (pwTarget1, pData, dData);
            free    (pwTarget1);
            }
        else
            {
            printf (L"\r\nUnable to save the allocation bits\r\n");
            }
        }
    if (dOptions & DISPLAY_OPTION_ROOT)
        {
        if ((pwTarget1 = PdbTarget (pwPath, pwTarget,
                                    DISPLAY_OPTION_ROOT))
            != NULL)
            {
            PdbSave (pwTarget1, ppr, dRoot);
            free    (pwTarget1);
            }
        else
            {
            printf (L"\r\nUnable to save the root stream\r\n");
            }
        }
    if (dOptions & DISPLAY_OPTION_DATA)
        {
        for (dStream = 0; dStream < ppr->wCount; dStream++)
            {
            pwTarget1 = PdbTarget (pwPath, pwTarget,
                                   DISPLAY_OPTION_DATA | dStream);

            pData = PdbStream (pph, ppr, dStream, &dData);

            if ((pwTarget1 != NULL) && (pData != NULL))
                {
                PdbSave (pwTarget1, pData, dData);
                }
            else
                {
                printf (L"\r\nUnable to save data stream #%lu\r\n",
                        dStream);
                }
            free (pwTarget1);
            free (pData);
            }
        }
    return;
    }

// -----------------------------------------------------------------

VOID PdbExplode (PWORD pwPath,
                 PWORD pwTarget,
                 DWORD dOptions)
    {
    PPDB_HEADER pph;
    PPDB_ROOT   ppr;
    DWORD       dRoot;
    BOOL        fInvalid;

    if ((pph = PdbLoad (pwPath, &fInvalid)) != NULL)
        {
        if ((ppr = PdbRoot (pph, &dRoot)) != NULL)
            {
            PdbInfo   (pwPath, pph, ppr);
            PdbOutput (pwPath, pwTarget, dOptions, pph, ppr, dRoot);

            free (ppr);
            }
        free (pph);
        }
    else
        {
        printf ((fInvalid
                 ? L"\r\n\"%s\" is not a PDB file\r\n"
                 : L"\r\nUnable to load \"%s\"\r\n"),
                pwPath);
        }
    return;
    }

// =================================================================
// OPTION MANAGEMENT
// =================================================================

VOID WINAPI Option (BOOL   fPlus,
                    PDWORD pdOptions,
                    DWORD  dMask,
                    DWORD  dValue,
                    DWORD  dDefault)
    {
    DWORD dNewValue = *pdOptions & dMask;

    *pdOptions &= ~dMask;

    if (fPlus)
        {
        dNewValue = dValue & dMask;
        }
    else
        {
        if (dNewValue == (dValue & dMask))
            {
            dNewValue = dDefault & dMask;
            }
        }
    *pdOptions |= dNewValue;
    return;
    }

// =================================================================
// MAIN PROGRAM
// =================================================================

DWORD Main (DWORD argc, PWORD *argv, PWORD *argp)
    {
    PWORD pwTarget = L"";
    DWORD dOptions, i, j;
    BOOL  fPlus;

    printf (atAbout);

    if (argc < 2)
        {
        printf (atUsage, awArguments);
        printf (awMoreInfo);
        }
    else
        {
        dOptions = DISPLAY_OPTION_DEFAULT;

        for (i = 1; i < argc; i++)
            {
            switch (argv [i] [0])
                {
                default:
                    {
                    PdbExplode (argv [i], pwTarget, dOptions);
                    break;
                    }
                case '/':
                    {
                    pwTarget = argv [i] + 1;
                    break;
                    }
                case '+':
                case '-':
                    {
                    for (j = 0; argv [i] [j]; j++)
                        {
                        switch (argv [i] [j])
                            {
                            case '+':
                                {
                                fPlus = TRUE;
                                break;
                                }
                            case '-':
                                {
                                fPlus = FALSE;
                                break;
                                }
                            case 'h':
                            case 'H':
                                {
                                Option (fPlus, &dOptions,
                                        DISPLAY_OPTION_HEADER,
                                        DISPLAY_OPTION_HEADER,
                                        DISPLAY_OPTION_DEFAULT);
                                break;
                                }
                            case 'a':
                            case 'A':
                                {
                                Option (fPlus, &dOptions,
                                        DISPLAY_OPTION_ALLOC,
                                        DISPLAY_OPTION_ALLOC,
                                        DISPLAY_OPTION_DEFAULT);
                                break;
                                }
                            case 'r':
                            case 'R':
                                {
                                Option (fPlus, &dOptions,
                                        DISPLAY_OPTION_ROOT,
                                        DISPLAY_OPTION_ROOT,
                                        DISPLAY_OPTION_DEFAULT);
                                break;
                                }
                            case 'd':
                            case 'D':
                                {
                                Option (fPlus, &dOptions,
                                        DISPLAY_OPTION_DATA,
                                        DISPLAY_OPTION_DATA,
                                        DISPLAY_OPTION_DEFAULT);
                                break;
                                }
                            }
                        }
                    break;
                    }
                }
            }
        }
    return 0;
    }

// =================================================================
// END OF PROGRAM
// =================================================================
